import argparse
import boto3
from . app import Awsume


def awsume(profile_name: str = None, *args: list, **kwargs: dict) -> boto3.Session:
    cli_arguments = list(args) if args is not None else []

    for key, value in kwargs.items():
        newkey = '--' + key.replace('_', '-')
        cli_arguments.append(str(newkey))
        if not isinstance(value, bool):
            cli_arguments.append(str(value))

    app = Awsume(is_interactive=False)
    return app.run([profile_name] + cli_arguments)



import os
import configparser
import json
from pathlib import Path
import boto3

REGIONS = [
    'us-east-2',
    'us-east-1',
    'us-west-1',
    'us-west-2',
    'ap-east-1',
    'ap-south-1',
    'ap-northeast-3',
    'ap-northeast-2',
    'ap-southeast-1',
    'ap-southeast-2',
    'ap-northeast-1',
    'ca-central-1',
    'cn-north-1',
    'cn-northwest-1',
    'eu-central-1',
    'eu-west-1',
    'eu-west-2',
    'eu-west-3',
    'eu-north-1',
    'me-south-1',
    'sa-east-1',
    'us-gov-east-1',
    'us-gov-west-1',
]


def get_aws_files() -> tuple:
    config_file = os.environ.get('AWS_CONFIG_FILE') if os.environ.get('AWS_CONFIG_FILE') else '~/.aws/config'
    credentials_file = os.environ.get('AWS_CREDENTIALS_FILE') if os.environ.get('AWS_CREDENTIALS_FILE') else '~/.aws/credentials'
    return str(Path(config_file).expanduser()), str(Path(credentials_file).expanduser())


def get_profile_names(credentials_file: str, config_file: str) -> list:
    credentials = configparser.ConfigParser()
    credentials.read(credentials_file)
    profiles = list(credentials._sections.keys())

    config = configparser.ConfigParser()
    config.read(config_file)
    config_profiles = [_.replace('profile ', '') for _ in list(config._sections.keys())]

    return uniquely_concat_lists(profiles, config_profiles)


def uniquely_concat_lists(list1, list2):
    for element in list2:
        if element not in list1:
            list1.append(element)
    return list1


def profile_name_completer(prefix, parsed_args, **kwargs):
    config, credentials = get_aws_files()
    profile_names = get_profile_names(credentials, config)
    autocomplete_file = str(Path('~/.awsume/autocomplete.json').expanduser())
    if os.path.isfile(autocomplete_file):
        autocomplete = json.load(open(autocomplete_file))
        profile_names = uniquely_concat_lists(profile_names, autocomplete['profile-names'])
    return [profile_name for profile_name in profile_names if profile_name.startswith(prefix)]


def region_completer(prefix, parsed_args, **kwargs):
    return [region for region in REGIONS if region.startswith(prefix)]



version = '4.4.1'

name = 'awsume'
author = 'Trek10, Inc'
author_email = 'package-management@trek10.com'
description = 'Awsume - A cli that makes using AWS IAM credentials easy'
license = 'MIT'
homepage = 'https://github.com/trek10inc/awsume'

message = 'Thank you for using AWSume! Check us out at https://trek10.com'



import argparse
import json
import re
import operator
from datetime import datetime
import dateutil
import colorama
import difflib
from . import exceptions
from . logger import logger
from . safe_print import safe_print
from . import aws as aws_lib
from collections import OrderedDict
from difflib import SequenceMatcher

VALID_CREDENTIAL_SOURCES = [ None, 'Environment', 'Ec2InstanceMetadata', 'EcsContainer' ]
try:
    import Levenshtein
except:
    Levenshtein = False


def parse_time(date_time: datetime):
    date_time.replace(tzinfo=dateutil.tz.tzlocal())
    return date_time.strftime('%Y-%m-%d %H:%M:%S')


def profile_to_credentials(profile: dict) -> dict:
    if profile:
        return {
            'AccessKeyId': profile.get('aws_access_key_id'),
            'SecretAccessKey': profile.get('aws_secret_access_key'),
            'SessionToken': profile.get('aws_session_token'),
            'Region': profile.get('region'),
        }
    return {}


def credentials_to_profile(credentials: dict) -> dict:
    result =  {}
    if credentials.get('AccessKeyId'):
        result['aws_access_key_id'] = credentials['AccessKeyId']
    if credentials.get('SecretAccessKey'):
        result['aws_secret_access_key'] = credentials['SecretAccessKey']
    if credentials.get('SessionToken'):
        result['aws_session_token'] = credentials['SessionToken']
    if credentials.get('Region'):
        result['region'] = credentials['Region']
    return result


def validate_profile(config: dict, arguments: argparse.Namespace, profiles: dict, target_profile_name: str) -> bool:
    logger.debug('Validating profile')
    profile = profiles.get(target_profile_name)
    if arguments.output_profile and not is_mutable_profile(profiles, arguments.output_profile):
        raise exceptions.ImmutableProfileError(arguments.output_profile, 'not awsume-managed')
    if not profile:
        raise exceptions.ProfileNotFoundError(profile_name=target_profile_name)

    # validate role profiles
    if 'role_arn' in profile:
        if profile.get('credential_process'):
            raise exceptions.InvalidProfileError(target_profile_name, message='awsume does not support the credential_process profile option: {}')
        if profile.get('credential_source') and profile.get('source_profile'):
            raise exceptions.InvalidProfileError(target_profile_name, message='credential_source and source_profile are mutually exclusive profile options')
        if not profile.get('credential_source') and not profile.get('source_profile') and not profile.get('principal_arn'):
            raise exceptions.InvalidProfileError(target_profile_name, message='role profiles must contain one of credential_source or source_profile')
        if profile.get('credential_source') not in VALID_CREDENTIAL_SOURCES:
            raise exceptions.InvalidProfileError(target_profile_name, message='unsupported awsume credential_source profile option: {}'.format(profile.get('credential_source')))
        source_profile_name = profile.get('source_profile')
        source_profile = get_source_profile(profiles, target_profile_name)
        if source_profile_name and not source_profile:
            raise exceptions.ProfileNotFoundError(profile_name=source_profile_name)
        if source_profile and not source_profile.get('role_arn'):
            user_profile = source_profile
            user_profile_name = source_profile_name
        else:
            user_profile = None
            user_profile_name = None
    else:
        user_profile = profile
        user_profile_name = target_profile_name

    # validate user profile
    if user_profile:
        missing_keys = []
        if 'credential_source' in profile:
            if profile.get('credential_source') not in VALID_CREDENTIAL_SOURCES:
                raise exceptions.InvalidProfileError(user_profile_name, message='unsupported awsume credential_source profile option: {}'.format(profile.get('credential_source')))
        else:
            if 'aws_access_key_id' not in user_profile:
                missing_keys.append('aws_access_key_id')
            if 'aws_secret_access_key' not in user_profile:
                missing_keys.append('aws_secret_access_key')
        if missing_keys:
            raise exceptions.InvalidProfileError(user_profile_name, message='Missing keys {}, or credential_source'.format(', '.join(missing_keys)))

    # validate arguments with profile
    if 'role_arn' not in profile and arguments.auto_refresh:
        raise exceptions.ValidationException('Cannot use autoawsume with non-role profile')
    return True


def is_mutable_profile(profiles: dict, profile_name: str) -> dict:
    profile = profiles.get(profile_name)
    if not profile:
        return True
    if profile.get('manager') == 'awsume':
        return True
    return False


def get_source_profile(profiles: dict, target_profile_name: str) -> dict:
    target_profile = profiles.get(target_profile_name)
    if target_profile:
        source_profile_name = target_profile.get('source_profile', 'default')
        return profiles.get(source_profile_name)
    return None


def get_role_chain(profiles: dict, target_profile_name: str) -> list:
    target_profile = profiles.get(target_profile_name)
    if not target_profile or not target_profile.get('role_arn'):
        return [target_profile_name]

    role_chain = []
    while target_profile:
        if target_profile.get('role_arn'):
            if target_profile_name in role_chain:
                raise exceptions.InvalidProfileError(','.join(role_chain), 'cannot have circular role-chains')
            role_chain.append(target_profile_name)
        target_profile_name = target_profile.get('source_profile')
        target_profile = profiles.get(target_profile_name)
    role_chain.reverse()
    return role_chain


def get_region(profiles: dict, arguments: argparse.Namespace, config: dict, ignore_config: bool = False, ignore_default: bool = False) -> str:
    if arguments.region:
        return arguments.region
    if arguments.role_arn and arguments.source_profile and profiles.get(arguments.source_profile, {}).get('region'):
        return profiles[arguments.source_profile]['region']
    target_profile = profiles.get(arguments.target_profile_name, {})
    if target_profile.get('region'):
        return target_profile['region']
    source_profile = get_source_profile(profiles, arguments.target_profile_name)
    if source_profile and source_profile.get('region'):
        return source_profile['region']
    if not ignore_default and profiles.get('default', {}).get('region'):
        return profiles['default']['region']
    if not ignore_config and config.get('region'):
        return config['region']
    return None


def get_external_id(arguments: argparse.Namespace, target_profile: dict):
    if arguments.external_id:
        return arguments.external_id
    return target_profile.get('external_id')


def get_role_duration(config: dict, arguments: argparse.Namespace, target_profile: dict):
    if arguments.role_duration:
        return int(arguments.role_duration)
    if target_profile.get('duration_seconds'):
        return int(target_profile.get('duration_seconds'))
    if config.get('role-duration'):
        return int(config.get('role-duration'))
    return 0


def get_mfa_serial(profiles: dict, target_profile_name: str) -> dict:
    target_profile = profiles.get(target_profile_name)
    mfa_serial = target_profile.get('mfa_serial')
    if not mfa_serial:
        source_profile_name = target_profile.get('source_profile')
        mfa_serial = profiles.get(source_profile_name, {}).get('mfa_serial')
    return mfa_serial


def get_mfa_token() -> str:
    token_pattern = re.compile('^[0-9]{6}$')
    safe_print('Enter MFA token: ', colorama.Fore.CYAN, end='')
    while True:
        mfa_token = input()
        if token_pattern.match(mfa_token):
            return mfa_token
        else:
            safe_print('Please enter a valid MFA token: ', colorama.Fore.CYAN, end='')


def aggregate_profiles(result: list) -> dict:
    return_profiles = {}
    for profiles in result:
        for profile_name, profile in profiles.items():
            if profile_name not in return_profiles:
                return_profiles[profile_name] = profile
            else:
                return_profiles[profile_name].update(profile)
    return return_profiles


def format_aws_profiles(profiles: dict, get_extra_data: bool) -> list: # pragma: no cover
    sorted_profiles = OrderedDict(sorted(profiles.items()))
    # List headers
    list_headers = ['PROFILE', 'TYPE', 'SOURCE', 'MFA?', 'REGION', 'ACCOUNT']
    profile_list = []
    profile_list.append([])
    profile_list[0].extend(list_headers)
    # now fill the tables with the appropriate data
    for name in sorted_profiles:
        #don't add any autoawsume profiles
        if 'auto-refresh-' not in name:
            profile = sorted_profiles[name]
            is_role_profile = 'role_arn' in profile
            profile_type = 'Role' if is_role_profile else 'User'
            # Assume source_profile is None unless we find otherwise afterwards.
            if is_role_profile:
                if 'source_profile' in profile.keys():
                    source_profile = profile['source_profile']
                elif 'principal_arn' in profile.keys():
                    source_profile = profile['principal_arn'].split(':')[-1]
                else:
                    source_profile = 'None'
            else:
                source_profile = 'None'
            mfa_needed = 'Yes' if 'mfa_serial' in profile else 'No'
            profile_region = str(profile.get('region'))
            profile_account_id = get_account_id(profile, get_extra_data)
            list_row = [name, profile_type, source_profile, mfa_needed, profile_region, profile_account_id]
            profile_list.append(list_row)
    return profile_list


def print_formatted_data(profile_data: list): # pragma: no cover
    print('Listing...\n')
    widths = [max(map(len, col)) for col in zip(*profile_data)]
    print('AWS Profiles'.center(sum(widths) + 10, '='))
    for row in profile_data:
        print('  '.join((val.ljust(width) for val, width in zip(row, widths))))


def list_profile_data(profiles: dict, get_extra_data: bool): # pragma: no cover
    profiles = {k: v for k, v in profiles.items() if not v.get('autoawsume')}
    formatted_profiles = format_aws_profiles(profiles, get_extra_data)
    print_formatted_data(formatted_profiles)


def get_account_id(profile: dict, call_aws: bool = False) -> str:
    logger.info('Getting account ID from profile: %s', json.dumps(profile, indent=2))
    if profile.get('role_arn'):
        return profile['role_arn'].replace('arn:aws:iam::', '').split(':')[0]
    if profile.get('mfa_serial'):
        return profile['mfa_serial'].replace('arn:aws:iam::', '').split(':')[0]
    if call_aws and profile.get('aws_access_key_id') and profile.get('aws_secret_access_key'):
        return aws_lib.get_account_id(profile_to_credentials(profile))
    return 'Unavailable'


def get_profile_name(config: dict, profiles: dict, target_profile_name: str, log=True) -> str:
    if target_profile_name in profiles or not config.get('fuzzy-match'):
        return target_profile_name
    profile_names = list(profiles.keys())

    matched_prefix_profile = match_prefix(profile_names, target_profile_name)
    if matched_prefix_profile:
        if log:
            safe_print('Using profile ' + matched_prefix_profile, color=colorama.Fore.YELLOW)
        return matched_prefix_profile

    matched_contains_profile = match_contains(profile_names, target_profile_name)
    if matched_contains_profile:
        if log:
            safe_print('Using profile ' + matched_contains_profile, color=colorama.Fore.YELLOW)
        return matched_contains_profile

    matched_levenshtein_profile = match_levenshtein(profile_names, target_profile_name)
    if matched_levenshtein_profile:
        if log:
            safe_print('Using profile ' + matched_levenshtein_profile, color=colorama.Fore.YELLOW)
        return matched_levenshtein_profile

    return target_profile_name


def match_prefix(profile_names: list, profile_name: str) -> str:
    prefix_words = [_ for _ in profile_names if _.startswith(profile_name)]
    if len(prefix_words) == 1:
        return prefix_words[0]
    return None


def match_contains(profile_names: list, profile_name: str) -> str:
    def longest_contains(str1, str2):
        sequence_match = SequenceMatcher(None,str1,str2)
        match = sequence_match.find_longest_match(0, len(str1), 0, len(str2))
        return match.size

    matches = {profile: longest_contains(profile_name, profile) for profile in profile_names}
    biggest_match = max(matches.values())
    result = [k for k in matches.keys() if matches[k] == biggest_match]
    if len(result) == 1:
        return result[0]
    return None


def match_levenshtein(profile_names: list, profile_name: str) -> str:
    if not Levenshtein:
        logger.debug('Levenshtein not installed, try installing awsume[fuzzy]')
        return None
    matches = {profile: Levenshtein.distance(profile_name, profile) for profile in profile_names}
    closest_match = min(matches.values())
    result = [k for k in matches.keys() if matches[k] == closest_match]
    if len(result) == 1:
        return result[0]
    return None



import argparse
import configparser
import json
import os
import colorama


from . lib import exceptions
from . hookimpl import hookimpl
from .. import __data__
from ..autoawsume.process import kill
from . lib import aws as aws_lib
from . lib import aws_files as aws_files_lib
from . lib.logger import logger
from . lib.safe_print import safe_print
from . lib import config_management as config_lib
from . lib import profile as profile_lib
from . lib import cache as cache_lib
from . lib.profile import VALID_CREDENTIAL_SOURCES
from . lib.profile import get_role_chain, get_profile_name


def custom_duration_argument_type(string):
    number = int(string)
    if number >= 0 and number <= 43201:
        return number
    raise argparse.ArgumentTypeError('Custom Duration must be between 0 and 43200')


@hookimpl(tryfirst=True)
def add_arguments(config: dict, parser: argparse.ArgumentParser):
    logger.info('Adding arguments')
    parser.add_argument('-v', '--version',
        action='store_true',
        dest='version',
        help='Display the current version of awsume',
    )
    parser.add_argument('-o', '--output-profile',
        action='store',
        dest='output_profile',
        metavar='output_profile',
        help='A profile to output credentials to',
    )
    parser.add_argument('--clean',
        action='store_true',
        dest='clean',
        help='Clean expired output profiles',
    )
    parser.add_argument('profile_name',
        nargs='?',
        action='store',
        metavar='profile_name',
        help='The target profile name',
    )
    parser.add_argument('-r', '--refresh',
        action='store_true',
        dest='force_refresh',
        help='Force refresh credentials',
    )
    parser.add_argument('-s', '--show-commands',
        action='store_true',
        dest='show_commands',
        help='Show the commands to set the credentials',
    )
    parser.add_argument('-u', '--unset',
        action='store_true',
        dest='unset_variables',
        help='Unset your aws environment variables',
    )
    parser.add_argument('-a', '--auto-refresh',
        action='store_true',
        dest='auto_refresh',
        help='Auto refresh credentials',
    )
    parser.add_argument('-k', '--kill-refresher',
        action='store_true',
        default=False,
        dest='kill',
        help='Kill autoawsume',
    )
    parser.add_argument('-l', '--list-profiles',
        nargs='?',
        action='store',
        default=None,
        const='list',
        choices=['more', 'list', None],
        metavar='more',
        dest='list_profiles',
        help='List profiles, "more" for detail (slow)',
    )
    parser.add_argument('--refresh-autocomplete',
        action='store_true',
        dest='refresh_autocomplete',
        help='Refresh all plugin autocomplete profiles',
    )
    parser.add_argument('--role-arn',
        action='store',
        dest='role_arn',
        metavar='role_arn',
        help='Role ARN or <partition>:<account_id>:<role_name>',
    )
    parser.add_argument('--principal-arn',
        action='store',
        dest='principal_arn',
        metavar='principal_arn',
        help='Principal ARN or <partition>:<account_id>:<provider_name>',
    )
    parser.add_argument('--source-profile',
        action='store',
        dest='source_profile',
        metavar='source_profile',
        help='source_profile to use (role-arn only)',
    )
    parser.add_argument('--external-id',
        action='store',
        dest='external_id',
        metavar='external_id',
        help='External ID to pass to the assume_role',
    )
    parser.add_argument('--mfa-token',
        action='store',
        dest='mfa_token',
        metavar='mfa_token',
        help='Your mfa token',
    )
    parser.add_argument('--region',
        action='store',
        dest='region',
        metavar='region',
        help='The region you want to awsume into',
    )
    parser.add_argument('--session-name',
        action='store',
        dest='session_name',
        metavar='session_name',
        help='Set a custom role session name',
    )
    parser.add_argument('--role-duration',
        action='store',
        dest='role_duration',
        type=custom_duration_argument_type,
        metavar='role_duration',
        help='Seconds to get role creds for',
    )
    assume_role_method = parser.add_mutually_exclusive_group()
    assume_role_method.add_argument('--with-saml',
        action='store_true',
        dest='with_saml',
        help='Use saml (requires plugin)',
    )
    assume_role_method.add_argument('--with-web-identity',
        action='store_true',
        dest='with_web_identity',
        help='Use web identity (requires plugin)',
    )
    parser.add_argument('--json',
        action='store',
        dest='json',
        metavar='json',
        help='Use json credentials',
    )
    parser.add_argument('--credentials-file',
        action='store',
        dest='credentials_file',
        metavar='credentials_file',
        help='Target a shared credentials file',
    )
    parser.add_argument('--config-file',
        action='store',
        dest='config_file',
        metavar='config_file',
        help='Target a config file',
    )
    parser.add_argument('--config',
        nargs='*',
        dest='config',
        action='store',
        metavar='option',
        help='Configure awsume',
    )
    parser.add_argument('--list-plugins',
        action='store_true',
        dest='list_plugins',
        help='List installed plugins',
    )
    parser.add_argument('--info',
        action='store_true',
        dest='info',
        help='Print any info logs to stderr',
    )
    parser.add_argument('--debug',
        action='store_true',
        dest='debug',
        help='Print any debug logs to stderr',
    )


@hookimpl(tryfirst=True)
def post_add_arguments(config: dict, arguments: argparse.Namespace, parser: argparse.ArgumentParser):
    logger.debug('Post add arguments')
    logger.debug(json.dumps(vars(arguments)))
    if arguments.auto_refresh:
        if arguments.role_arn:
            raise exceptions.ValidationException('Cannot use autoawsume with a given role_arn')
        if arguments.json:
            raise exceptions.ValidationException('Cannot use autoawsume with json')
    if arguments.clean:
        _, credentials_file = aws_files_lib.get_aws_files(arguments, config)
        aws_files_lib.remove_expired_output_profiles(credentials_file)
        raise exceptions.EarlyExit()
    if arguments.version:
        logger.debug('Logging version')
        safe_print(__data__.version)
        raise exceptions.EarlyExit()
    if arguments.unset_variables:
        logger.debug('Unsetting environment variables')
        print('Unset', [])
        raise exceptions.EarlyExit()
    if type(arguments.config) is list:
        config_lib.handle_config(arguments.config)
        raise exceptions.EarlyExit()
    if arguments.kill:
        kill(arguments)
        raise exceptions.EarlyExit()

    if arguments.with_saml:
        if bool(arguments.role_arn) is not bool(arguments.principal_arn):
            parser.error('both or neither --principal-arn and --role-arn must be specified with saml')
    if not arguments.with_saml and arguments.principal_arn:
        parser.error('--principal-arn can only be specified with --with-saml')

    if arguments.role_arn and not arguments.role_arn.startswith('arn:'):
        logger.debug('Using short-hand role arn syntax')
        parts = arguments.role_arn.split(':')
        if len(parts) == 2:
            partition = 'aws'
            account_id = parts[0]
            role_name = parts[1]
        elif len(parts) == 3:
            partition = parts[0]
            account_id = parts[1]
            role_name = parts[2]
        else:
            parser.error('--role-arn must be a valid role arn or follow the format "<partition>:<account_id>:<role_name>"')
        if not account_id.isnumeric() or len(account_id) != 12:
            parser.error('--role-arn account id must be valid numeric account id of length 12')
        arguments.role_arn = 'arn:{}:iam::{}:role/{}'.format(partition, account_id, role_name)

    if arguments.principal_arn and not arguments.principal_arn.startswith('arn:'):
        logger.debug('Using short-hand role arn syntax')
        parts = arguments.principal_arn.split(':')
        if len(parts) == 2:
            partition = 'aws'
            account_id = parts[0]
            provider_name = parts[1]
        elif len(parts) == 3:
            partition = parts[0]
            account_id = parts[1]
            provider_name = parts[2]
        else:
            parser.error('--principal-arn must be a valid role arn or follow the format "<partition>:<account_id>:<provider_name>"')
        if not provider_name.isnumeric() or len(provider_name) != 12:
            parser.error('--principal-arn account id must be valid numeric account id of length 12')
        arguments.principal_arn = 'arn:{}:iam::{}:role/{}'.format(partition, account_id, provider_name)

    if not arguments.profile_name:
        if arguments.role_arn:
            logger.debug('Role arn passed, target profile name will be role_arn')
            arguments.target_profile_name = arguments.role_arn
        else:
            logger.debug('No profile name passed, target profile name will be "default"')
            arguments.target_profile_name = 'default'
    else:
        arguments.target_profile_name = arguments.profile_name


@hookimpl(tryfirst=True)
def collect_aws_profiles(config: dict, arguments: argparse.Namespace, credentials_file: str, config_file: str):
    logger.info('Collecting AWS profiles')
    profiles = aws_files_lib.read_aws_file(credentials_file)
    config_profiles = aws_files_lib.read_aws_file(config_file)

    for profile_name, profile in config_profiles.items():
        short_name = profile_name.replace('profile ', '')
        if short_name not in profiles:
            profiles[short_name] = {}
        profiles[short_name].update(profile)
    logger.debug('Collected {} profiles'.format(len(profiles)))
    return profiles


@hookimpl(tryfirst=True)
def post_collect_aws_profiles(config: dict, arguments: argparse.Namespace, profiles: dict):
    logger.info('Post collect AWS profiles')
    if arguments.list_profiles:
        logger.debug('Listing profiles')
        profile_lib.list_profile_data(profiles, arguments.list_profiles == 'more')
        raise exceptions.EarlyExit()


def assume_role_from_cli(config: dict, arguments: dict, profiles: dict):
    region = profile_lib.get_region(profiles, arguments, config, ignore_config=True, ignore_default=True)
    logger.info('Using role_arn from the CLI')
    role_duration = arguments.role_duration or int(config.get('role-duration', 0))
    session_name = arguments.session_name or 'awsume-cli-role'
    logger.debug('Session name: {}'.format(session_name))
    if not arguments.source_profile:
        logger.debug('Using current credentials to assume role')
        role_session = aws_lib.assume_role({}, arguments.role_arn, session_name, region=region, external_id=arguments.external_id, role_duration=role_duration)
    else:
        logger.debug('Using the source_profile from the cli to call assume_role')
        source_profile = profiles.get(arguments.source_profile)
        if not source_profile:
            raise exceptions.ProfileNotFoundError(profile_name=arguments.source_profile)
        source_credentials = profile_lib.profile_to_credentials(source_profile)
        mfa_serial = source_profile.get('mfa_serial')
        if role_duration:
            logger.debug('Using custom role duration')
            if mfa_serial:
                logger.debug('Requires MFA')
                logger.debug('Using custom role duration for role that needs mfa_serial, skipping get-session-token call')
                source_session = source_credentials
                role_session = aws_lib.assume_role(
                    source_session,
                    arguments.role_arn,
                    session_name,
                    region=region,
                    external_id=arguments.external_id,
                    role_duration=role_duration,
                    mfa_serial=mfa_serial,
                    mfa_token=arguments.mfa_token,
                )
            else:
                logger.debug('MFA not needed, assuming role from with profile creds')
                role_session = aws_lib.assume_role(
                    source_credentials,
                    arguments.role_arn,
                    session_name,
                    region=region,
                    external_id=arguments.external_id,
                    role_duration=role_duration,
                )
        else:
            logger.debug('Using default role duration')
            if mfa_serial:
                logger.debug('MFA required')
                source_session = aws_lib.get_session_token(
                    source_credentials,
                    region=profile_lib.get_region(profiles, arguments, config),
                    mfa_serial=mfa_serial,
                    mfa_token=arguments.mfa_token,
                    ignore_cache=arguments.force_refresh,
                    duration_seconds=config.get('debug', {}).get('session_token_duration'),
                )
            else:
                logger.debug('MFA not required')
                source_session = source_credentials
            role_session = aws_lib.assume_role(
                source_session,
                arguments.role_arn,
                session_name,
                region=region,
                external_id=arguments.external_id,
                role_duration=role_duration,
            )
    return role_session


def get_assume_role_credentials(config: dict, arguments: argparse.Namespace, profiles: dict, target_profile: dict, role_duration: int, source_credentials: dict, target_profile_name: str):
    logger.info('Getting assume role credentials')
    region = profile_lib.get_region(profiles, arguments, config)
    external_id = profile_lib.get_external_id(arguments, target_profile)
    if not source_credentials:
        source_profile = profile_lib.get_source_profile(profiles, target_profile_name)
        source_credentials = profile_lib.profile_to_credentials(source_profile)
    role_session = aws_lib.assume_role(
        source_credentials,
        target_profile.get('role_arn'),
        arguments.session_name or target_profile_name,
        region=region,
        external_id=external_id,
        role_duration=role_duration,
    )
    if 'SourceExpiration' in source_credentials:
        role_session['SourceExpiration'] = source_credentials['SourceExpiration']
    elif 'Expiration' in source_credentials:
        role_session['SourceExpiration'] = source_credentials['Expiration']
    return role_session


def get_assume_role_credentials_mfa_required(config: dict, arguments: argparse.Namespace, profiles: dict, target_profile: dict, role_duration: int, source_credentials: dict, target_profile_name: str):
    logger.info('Getting assume role credentials MFA required')
    region = profile_lib.get_region(profiles, arguments, config)
    mfa_serial = profile_lib.get_mfa_serial(profiles, target_profile_name)
    external_id = profile_lib.get_external_id(arguments, target_profile)
    source_profile = profile_lib.get_source_profile(profiles, target_profile_name)
    if source_profile:
        if 'role_arn' not in source_profile:
            logger.debug('Calling get_session_token to assume role with')
            if not source_credentials:
                source_credentials = profile_lib.profile_to_credentials(source_profile)
            source_session = aws_lib.get_session_token(
                source_credentials,
                region=region,
                mfa_serial=mfa_serial,
                mfa_token=arguments.mfa_token,
                ignore_cache=arguments.force_refresh,
                duration_seconds=config.get('debug', {}).get('session_token_duration'),
            )
        else:
            source_session = source_credentials
    elif 'credential_source' in target_profile and target_profile['credential_source'] in VALID_CREDENTIAL_SOURCES:
        logger.debug('Using current environment to assume role')
        source_session = {}

    if arguments.auto_refresh and (os.environ.get('AWS_PROFILE', '').startswith('autoawsume-') or profiles.get(os.getenv('AWS_PROFILE'), {}).get('autoawsume')):
        os.environ.pop('AWS_PROFILE')
        os.environ.pop('AWS_DEFAULT_PROFILE')

    role_session = aws_lib.assume_role(
        source_session,
        target_profile.get('role_arn'),
        arguments.session_name or target_profile_name,
        region=region,
        external_id=external_id,
        role_duration=role_duration,
    )

    if 'SourceExpiration' in source_session:
        role_session['SourceExpiration'] = source_session['SourceExpiration']
    elif 'Expiration' in source_session:
        role_session['SourceExpiration'] = source_session['Expiration']

    return source_session, role_session


def get_assume_role_credentials_mfa_required_large_custom_duration(config: dict, arguments: argparse.Namespace, profiles: dict, target_profile: dict, role_duration: int, target_profile_name: str):
    logger.info('Getting assume role credentials MFA required, large custom duration')
    if arguments.auto_refresh and role_duration > 3600:
        raise exceptions.ValidationException('Cannot use autoawsume with custom role duration of more than 1 hour')
    logger.debug('Skipping the get_session_token call, temp creds cannot be used for custom role duration')

    region = profile_lib.get_region(profiles, arguments, config)
    mfa_serial = profile_lib.get_mfa_serial(profiles, target_profile_name)
    external_id = profile_lib.get_external_id(arguments, target_profile)
    source_profile = profile_lib.get_source_profile(profiles, target_profile_name)
    source_session = profile_lib.profile_to_credentials(source_profile)

    role_session = aws_lib.assume_role(
        source_session,
        target_profile.get('role_arn'),
        arguments.session_name or target_profile_name,
        region=region,
        external_id=external_id,
        role_duration=role_duration,
        mfa_serial=mfa_serial,
        mfa_token=arguments.mfa_token,
    )
    return role_session


def get_credentials_no_mfa(config: dict, arguments: argparse.Namespace, profiles: dict, target_profile: dict):
    logger.info('Getting credentials MFA not required')
    region = profile_lib.get_region(profiles, arguments, config)
    return_session = profile_lib.profile_to_credentials(target_profile)
    return_session['Region'] = region
    return return_session


def get_credentials_from_credential_source(config: dict, arguments: argparse.Namespace, profiles: dict, target_profile: dict, target_profile_name: str):
    logger.info('Getting credentials from credential_source')
    region = profile_lib.get_region(profiles, arguments, config)
    return_session = {'AwsProfile': target_profile_name}
    return_session['Region'] = region
    return return_session


def get_session_token_credentials(config: dict, arguments: argparse.Namespace, profiles: dict, target_profile: dict, target_profile_name: str):
    logger.info('Getting session token credentials')
    region = profile_lib.get_region(profiles, arguments, config)
    mfa_serial = profile_lib.get_mfa_serial(profiles, target_profile_name)
    source_credentials = profile_lib.profile_to_credentials(target_profile)
    user_session = aws_lib.get_session_token(
        source_credentials,
        region=region,
        mfa_serial=mfa_serial,
        mfa_token=arguments.mfa_token,
        ignore_cache=arguments.force_refresh,
        duration_seconds=config.get('debug', {}).get('session_token_duration'),
    )
    return user_session


def get_credentials_handler(config: dict, arguments: argparse.Namespace, profiles: dict, profile_name: str, credentials: dict) -> dict:
    credentials = credentials if credentials else {}
    logger.info('Getting credentials')

    user_session = None
    role_session = None

    if arguments.role_arn:
        role_session = assume_role_from_cli(config, arguments, profiles)
    else:
        profile_lib.validate_profile(config, arguments, profiles, profile_name)
        target_profile = profiles.get(profile_name)
        mfa_serial = profile_lib.get_mfa_serial(profiles, profile_name)
        role_duration = profile_lib.get_role_duration(config, arguments, target_profile)

        if 'role_arn' in target_profile:
            logger.debug('assume_role call needed')
            if mfa_serial and not credentials: # if using specific credentials, no mfa needed
                if role_duration > 3600: # cannot use temp creds with custom role duration more than an hour
                    role_session = get_assume_role_credentials_mfa_required_large_custom_duration(config, arguments, profiles, target_profile, role_duration, profile_name)
                else:
                    user_session, role_session = get_assume_role_credentials_mfa_required(config, arguments, profiles, target_profile, role_duration, credentials, profile_name)
            else:
                role_session = get_assume_role_credentials(config, arguments, profiles, target_profile, role_duration, credentials, profile_name)
        else:
            if mfa_serial:
                user_session = get_session_token_credentials(config, arguments, profiles, target_profile, profile_name)
            elif 'credential_source' in target_profile:
                user_session = get_credentials_from_credential_source(config, arguments, profiles, target_profile, profile_name)
            else:
                user_session = get_credentials_no_mfa(config, arguments, profiles, target_profile)

    if config.get('is_interactive'):
        if user_session and user_session.get('Expiration'):
            safe_print('Session token will expire at {}'.format(profile_lib.parse_time(user_session['Expiration'])), colorama.Fore.GREEN)
        if role_session and role_session.get('Expiration'):
            safe_print('[{}] Role credentials will expire {}'.format(profile_name, profile_lib.parse_time(role_session['Expiration'])), colorama.Fore.GREEN)

    return role_session or user_session


@hookimpl(tryfirst=True)
def get_credentials(config: dict, arguments: argparse.Namespace, profiles: dict) -> dict:
    if arguments.role_arn:
        target_profile_name = arguments.role_arn
    else:
        target_profile_name = get_profile_name(config, profiles, arguments.target_profile_name)
    role_chain = get_role_chain(profiles, target_profile_name)
    credentials = None
    for profile_name in role_chain:
        credentials = get_credentials_handler(config=config, arguments=arguments, profiles=profiles, profile_name=profile_name, credentials=credentials)
    return credentials



import sys
import logging
import re


class LogFormatter(logging.Formatter):
    @staticmethod
    def _filter(s):
        no_access_key_id = re.sub(r'(?<![A-Z0-9])[A-Z0-9]{20}(?![A-Z0-9])', 'SECRET', s)
        no_secret_access_key = re.sub(r'(?<![A-Za-z0-9/+=])[A-Za-z0-9/+=]{40}(?![A-Za-z0-9/+=])', 'SECRET', no_access_key_id)
        return no_secret_access_key

    def format(self, record):
        original = logging.Formatter.format(self, record)
        return self._filter(original)


logger = logging.getLogger('awsume') # type: logging.Logger
LOG_HANDLER = logging.StreamHandler()
LOG_HANDLER.setFormatter(LogFormatter('[%(asctime)s] %(filename)s:%(funcName)s : [%(levelname)s] %(message)s'))
logger.addHandler(LOG_HANDLER)



import os, pathlib

BASH_AUTOCOMPLETE_SCRIPT = """
_awsume() {
    local cur prev opts
    COMPREPLY=()
    cur="${COMP_WORDS[COMP_CWORD]}"
    prev="${COMP_WORDS[COMP_CWORD-1]}"
    opts=$(awsume-autocomplete)
    COMPREPLY=( $(compgen -W "${opts}" -- ${cur}) )
    return 0
}
complete -F _awsume awsume
"""

ZSH_AUTOCOMPLETE_SCRIPT = """
#Auto-Complete function for AWSume
fpath=(~/.awsume/zsh-autocomplete/ $fpath)
"""

ZSH_AUTOCOMPLETE_FUNCTION = """#compdef awsume
_arguments "*: :($(awsume-autocomplete))"
"""

POWERSHELL_AUTOCOMPLETE_SCRIPT = """
Register-ArgumentCompleter -Native -CommandName awsume -ScriptBlock {
    param($wordToComplete, $commandAst, $cursorPosition)
    $(awsume-autocomplete) |
    Where-Object { $_ -like "$wordToComplete*" } |
    Sort-Object |
    ForEach-Object {
        [System.Management.Automation.CompletionResult]::new($_, $_, 'ParameterValue', $_)
    }
}
"""

SCRIPTS = {
    'bash': BASH_AUTOCOMPLETE_SCRIPT,
    'zsh': ZSH_AUTOCOMPLETE_SCRIPT,
    'powershell': POWERSHELL_AUTOCOMPLETE_SCRIPT,
}

def main(shell: str, autocomplete_file: str):
    autocomplete_file = str(pathlib.Path(autocomplete_file).expanduser())
    autocomplete_script = SCRIPTS[shell]

    basedir = os.path.dirname(autocomplete_file)
    if basedir and not os.path.exists(basedir):
        os.makedirs(basedir)
    open(autocomplete_file, 'a').close()

    if autocomplete_script in open(autocomplete_file, 'r').read():
        print('Autocomplete script already in ' + autocomplete_file)
    else:
        with open(autocomplete_file, 'a') as f:
            f.write('\n#Auto-Complete function for AWSume')
            f.write(autocomplete_script)
        print('Wrote autocomplete script to ' + autocomplete_file)

    # install autocomplete function if zsh
    if shell == 'zsh':
        zsh_autocomplete_function_file = str(pathlib.Path('~/.awsume/zsh-autocomplete/_awsume').expanduser())
        basedir = os.path.dirname(zsh_autocomplete_function_file)
        if basedir and not os.path.exists(basedir):
            os.makedirs(basedir)
        if not os.path.isfile(zsh_autocomplete_function_file):
            open(zsh_autocomplete_function_file, 'w').close()

        if ZSH_AUTOCOMPLETE_FUNCTION in open(zsh_autocomplete_function_file, 'r').read():
            print('Zsh function already in ' + zsh_autocomplete_function_file)
        else:
            with open(zsh_autocomplete_function_file, 'a') as f:
                f.write(ZSH_AUTOCOMPLETE_FUNCTION)
            print('Wrote zsh function to ' + zsh_autocomplete_function_file)



from . config_management import CONFIG_MANAGEMENT_HELP


class AwsumeException(Exception):
    """"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ProfileNotFoundError(AwsumeException):
    """"""
    def __init__(self, profile_name='', message=''):
        self.profile_name = profile_name
        self.message = message
    def __str__(self):
        if self.message:
            return self.message
        return 'Profile {} not found.'.format(self.profile_name)


class InvalidProfileError(AwsumeException):
    """"""
    def __init__(self, profile_name, message=''):
        self.profile_name = profile_name
        self.message = message
    def __str__(self):
        return 'Invalid profile [{}] {}'.format(self.profile_name, self.message)


class ImmutableProfileError(AwsumeException):
    """"""
    def __init__(self, profile_name, message=''):
        self.profile_name = profile_name
        self.message = message
    def __str__(self):
        return 'Immutable profile [{}] {}'.format(self.profile_name, self.message)


class ValidationException(AwsumeException):
    """"""
    def __init__(self, message='no message'):
        self.message = message
    def __str__(self):
        return 'Validation Exception - {}'.format(self.message)


class ConfigParseException(AwsumeException):
    """"""
    def __init__(self, file_name, message='no message', error=None):
        self.file_name = file_name
        self.message = message
        self.error = error
    def __str__(self):
        return 'ConfigParseException - ({}) {} {}'.format(self.file_name, self.message, self.error)


class ConfigOperationException(AwsumeException):
    """"""
    def __init__(self, message='no message'):
        self.message = message
    def __str__(self):
        return '{}\n{}'.format(self.message, CONFIG_MANAGEMENT_HELP)


class UserAuthenticationError(AwsumeException):
    """"""
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message if self.message else 'Unable to get session token'


class RoleAuthenticationError(AwsumeException):
    """"""
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message if self.message else 'Unable to assume role'


class SAMLAssertionNotFoundError(AwsumeException):
    """"""
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message if self.message else 'No SAML assertion'


class SAMLAssertionMissingRoleError(AwsumeException):
    """"""
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message if self.message else 'No role in the SAML assertion'


class SAMLRoleNotFoundError(AwsumeException):
    """"""
    def __init__(self, principal_arn, role_arn, message=''):
        self.role_arn = role_arn
        self.principal_arn = principal_arn
        self.message = message
    def __str__(self):
        return self.message if self.message else 'No match for SAML principal and role: {},{}'.format(self.principal_arn, self.role_arn)


class SAMLAssertionParseError(AwsumeException):
    """"""
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message if self.message else 'Cannot parse SAML assertion'


class NoCredentialsError(AwsumeException):
    """"""
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message if self.message else 'No credentials'


class EarlyExit(AwsumeException):
    """"""
    def __init__(self):
        pass
    def __str__(self):
        return 'Early exit exception, nothing left to do'



import base64
import json

try:
    import xmltodict
except:
    xmltodict = False

import colorama
from . safe_print import safe_print
from . exceptions import SAMLAssertionParseError, ValidationException


def parse_assertion(assertion: str) -> list:
    if not xmltodict:
        raise ValidationException(message='SAML option not installed, try installing awsume[saml]')

    roles = []
    response = xmltodict.parse(base64.b64decode(assertion))

    if response.get('saml2p:Response') is not None:
        attributes = response.get('saml2p:Response', {}).get('saml2:Assertion', {}).get('saml2:AttributeStatement', {}).get('saml2:Attribute', {})
        attribute_value_key = 'saml2:AttributeValue'
    else:
        attributes = response.get('samlp:Response', {}).get('saml:Assertion', {}).get('saml:AttributeStatement', {}).get('saml:Attribute', {})
        attribute_value_key = 'saml:AttributeValue'

    if not attributes:
        raise SAMLAssertionParseError()

    for attribute in [_ for _ in attributes if _.get('@Name', '') == 'https://aws.amazon.com/SAML/Attributes/Role']:
        if isinstance(attribute[attribute_value_key], list):
            for value in attribute[attribute_value_key]:
                roles.append(value['#text'])
        else:
            value = attribute[attribute_value_key]
            roles.append(value['#text'])

    return roles



import fastentrypoints
from setuptools import setup, find_packages

import awsume
from awsume.configure.post_install import CustomInstall

setup(
    name=awsume.__NAME__,
    packages=find_packages(),
    version=awsume.__VERSION__,
    author=awsume.__AUTHOR__,
    author_email=awsume.__AUTHOR_EMAIL__,
    description=awsume.__DESCRIPTION__,
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    license=awsume.__LICENSE__,
    url=awsume.__HOMEPAGE__,
    install_requires=[
        'colorama',
        'boto3',
        'psutil',
        'pluggy',
        'pyyaml',
    ],
    extras_require={
        'saml': ['xmltodict'],
        'fuzzy': ['python-levenshtein'],
    },
    scripts=[
        'shell_scripts/awsume',
        'shell_scripts/awsume.ps1',
        'shell_scripts/awsume.bat',
    ],
    entry_points={
        'console_scripts': [
            'awsumepy=awsume.awsumepy.main:main',
            'autoawsume=awsume.autoawsume.main:main',
            'awsume-configure=awsume.configure.main:main',
            'awsume-autocomplete=awsume_autocomplete:main',
        ],
    },
    python_requires='>=3.5',
    cmdclass={
        'install': CustomInstall,
    },
)



import json
import subprocess
import configparser
import time
import sys
import os
import logging
from pathlib import Path
from datetime import datetime, timedelta
import dateutil

from ..awsumepy.lib.aws_files import get_aws_files, delete_section
from ..awsumepy.lib.logger import LogFormatter
from ..awsumepy.lib.logger import logger as awsume_logger
from ..awsumepy.lib import exceptions
from .. import awsumepy

logger = logging.getLogger('autoawsume') # type: logging.Logger


def main():
    configure_logger()

    logger.debug('Getting credentials file')
    _, credentials_file = get_aws_files(None, None)
    logger.debug('Credentials file: {}'.format(credentials_file))

    while True:
        logger.info('Scanning profiles')
        credentials = configparser.ConfigParser()
        credentials.read(credentials_file)
        auto_profiles = {k: dict(v) for k, v in credentials._sections.items() if v.get('autoawsume')}

        expirations = []
        for profile_name, auto_profile in auto_profiles.items():
            logger.info('Looking at profile [{}]: {}'.format(profile_name, redact_profile(auto_profile)))
            expiration = datetime.strptime(auto_profile['expiration'], '%Y-%m-%d %H:%M:%S')
            source_expiration = datetime.strptime(auto_profile['source_expiration'], '%Y-%m-%d %H:%M:%S')
            if source_expiration < datetime.now():
                logger.debug('Source is expired')
                if expiration < datetime.now():
                    logger.debug('Role credentials are expired')
                    delete_profile(profile_name, credentials_file)
                else:
                    logger.debug('Role credentials are not expired')
                    expirations.append(expiration)
            else:
                logger.debug('Source credentials are not expired')
                if expiration - timedelta(seconds=60) < datetime.now():
                    logger.debug('Role credentials are expired or will expire in less than 60s')
                    session = refresh_profile(auto_profile)
                    if session:
                        logger.debug('Received session from awsume call')
                        expirations.append(session.awsume_credentials.get('Expiration'))
                    else:
                        logger.debug('No session returned from awsume call')
                        delete_profile(profile_name, credentials_file)
                else:
                    logger.debug('Role credentials are not expired')
                    expirations.append(expiration)
                    expirations.append(source_expiration)
        logger.debug('Collected expirations: {}'.format(json.dumps(expirations, default=str)))

        if not expirations:
            break

        local_expirations = [_.astimezone(dateutil.tz.tzlocal()) for _ in expirations]
        earliest_expiration = min(local_expirations)
        logger.debug('Earliest expiration: {}'.format(earliest_expiration))
        time_to_sleep = max(0, (earliest_expiration - datetime.now().replace(tzinfo=earliest_expiration.tzinfo)).total_seconds() - 60)
        logger.debug('Time to sleep: {}'.format(time_to_sleep))
        time.sleep(time_to_sleep)

    logger.info('Finished autoawsume')


def configure_logger():
    log_dir = str(Path('~/.awsume/logs/').expanduser())
    log_file = str(Path('~/.awsume/logs/autoawsume-{}'.format(datetime.now().strftime('%Y-%m-%d_%H:%M:%S:%f'))).expanduser())

    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    if not os.path.isfile(log_file):
        open(log_file, 'a').close()

    log_handler = logging.FileHandler(log_file)
    log_handler.setFormatter(LogFormatter('%(asctime)s | %(name)s | %(filename)s:%(funcName)s | [%(levelname)s] | %(message)s'))
    logger.addHandler(log_handler)
    logger.setLevel(logging.DEBUG)
    awsume_logger.handlers.clear()
    awsume_logger.addHandler(log_handler)
    awsume_logger.setLevel(logging.DEBUG)


def redact_profile(profile):
    redacted = { **profile }
    if 'aws_access_key_id' in redacted:
        redacted['aws_access_key_id'] = 'SECRET'
    if 'aws_secret_access_key' in redacted:
        redacted['aws_secret_access_key'] = 'SECRET'
    if 'aws_session_token' in redacted:
        redacted['aws_session_token'] = 'SECRET'
    return json.dumps(redacted, default=str)


def refresh_profile(auto_profile):
    logger.debug('Refreshing profile {}'.format(json.dumps(auto_profile, default=str)))
    try:
        session = awsumepy.awsume(*auto_profile.get('awsumepy_command').split(' '))
        logger.debug('Refreshed profile, returning session')
        return session
    except exceptions.AwsumeException as e:
        logger.debug('There was an issue refreshing the profile, not returning a session: {}'.format(e))
        logger.debug('', exc_info=True)
        return None


def delete_profile(profile, credentials):
    logger.info('Deleting profile [{}] from file: {}'.format(profile, credentials))
    # delete_section(profile, credentials)
    config = configparser.ConfigParser()

    logger.debug('Reading profiles')
    with open(str(credentials)) as f:
        config.read_file(f)
    logger.debug('Read profiles: {}'.format(list(config.keys())))

    if config.has_section(profile):
        logger.debug('Profile exists, removing it')
        config.remove_section(profile)

    logger.debug('Saving profiles: {}'.format(list(config.keys())))
    with open(str(credentials), 'w') as f:
        config.write(f)
    logger.debug('Saved changes')



from . import aws_files
from . import aws
from . import cache
from . import config_management
from . import constants
from . import exceptions
from . import logger
from . import profile
from . safe_print import safe_print